/* 
 * $Id: rib.h,v 1.9 1998/06/03 21:36:55 masaki Exp $
 */

#ifndef _RIB_H
#define _RIB_H

#include <stack.h>
#include <trace.h>
#include <hash.h>

#include <proto.h>
#include <interface.h>
#include <aspath.h>

#ifndef HAVE_PTHREAD_H
#include <pthread_fake.h>
#else
#include <pthread.h>
#endif /* HAVE_PTHREAD_H */

/*
 * "State" of routing table entry.
 */
#define	RTS_ACTIVE		0x0100	/* Route is active */
#define RTS_INTERIOR    	0x08	/* an interior route */
#define RTS_EXTERIOR    	0x10	/* an exterior route */
#define	RTS_HOLDDOWN		0x20	/* Route is held down */


#define	RTS_DELETE		0x40	/* Route is deleted */
#define	RTS_FREE		0x80	/* Route memory scheduled for free */


#define CONNECTED_PREF		0
#define STATIC_PREF		1
#define BGP_PREF		20
#define EBGP_PREF		20
#define OSPF_PREF		110
#define RIP_PREF		120
#define RIPNG_PREF		120
#define IBGP_PREF		200
#define KERNEL_PREF		250
#define HOLD_PREF		255
#define NO_PREF			255

#define RIB_UPD       0
#define RIB_DEL       2
/* #define RIB_HOLD      3 */
/* #define RIB_UNHOLD    4 */

#define KERNEL_ROUTE_ADD       0
#define KERNEL_ROUTE_CHG       1
#define KERNEL_ROUTE_DEL       2


#define MRT_RTOPT_SUPPRESS      0x01
#define MRT_RTOPT_AGGREGATE     0x02
#define MRT_RTOPT_NOINSTALL     0x04

struct _route_head_t;
typedef struct _route_node_t {
    struct _route_head_t *route_head;	/* pointer back to our route_head */
    int state;			/* mask of state flags */
    int pref;			/* ripe181 - as_in preference */
    int protocol_mask;		/* bit mask of protocols using this route */
    generic_attr_t *attr;
    time_t time;		/* time last updated (as in RIP) */
    u_long option;
} route_node_t;


typedef struct _route_head_t {
    prefix_t *prefix;
    route_node_t *active;	/* a pointer to best, or active route_node */
    LINKED_LIST *ll_route_nodes;
    u_long change_flag;
} route_head_t;


/*
 * Change is routing table entry
 */
#define	RTCF_NEXTHOP	0x01	/* Next hop change */
#define	RTCF_METRIC	0x02	/* Metric change */
#define	RTCF_METRIC2	0x04	/* Metric change */
#define	RTCF_ASPATH	0x08	/* AS path change */
#define	RTCF_TAG	0x10	/* Tag change */

typedef int (*proto_update_route_t) (prefix_t *prefix, generic_attr_t *new,
                         generic_attr_t *old, int pref);
typedef int (*kernel_update_route_t) (int cmd, prefix_t *dest,
                         prefix_t *next_hop, prefix_t *old_hop, int index, int oldindex);
typedef int (*rib_update_route_t) (int cmd, prefix_t *prefix, 
			 generic_attr_t *attr, int pref, u_long option);


typedef struct _rib_t {
    pthread_mutex_t mutex_lock;
    HASH_TABLE *hash;
    trace_t *trace;
    LINKED_LIST *ll_changed_route_nodes;
    proto_update_route_t proto_update_route[PROTO_MAX];
    kernel_update_route_t kernel_update_route;

#ifdef notdef
    int lock;		/* table is LOCKED -- do not modify while set */
    int num_active_routes; */
    LINKED_LIST *ll_route_nodes; /* delete this ? */
    void *user;		/* hook for program or user specific data */
#endif
} rib_t;

/* types returned by ret in rm_insert on addition to radix tree */
enum RM_RETURN_TYPES {
    RM_ROUTE_ALREADY_EXISTS,
    RM_ROUTE_LOWER_PREF,
    RM_ROUTE_BEST_PREF,
    RM_ROUTE_SAME_PREF,
    RM_NO_ACTIVE_ROUTE
};


rib_t * New_Rib (int maxbitlen);
void rib_open (rib_t * rib);
void rib_close (rib_t * rib);

char *proto2string (int proto);
generic_attr_t *New_Generic_Attr (int proto);
generic_attr_t *Ref_Generic_Attr (generic_attr_t * attr);
void Deref_Generic_Attr (generic_attr_t * attr);

route_node_t *rib_add_route (rib_t *rib, prefix_t *prefix, 
		             generic_attr_t *attr, int pref, u_long option);
int rib_del_route (rib_t * rib, prefix_t * prefix, generic_attr_t * attr);

void rib_show_route_head (uii_connection_t *uii, char *append);

int show_rib_status (uii_connection_t *uii);
void rib_show_route_line (uii_connection_t *uii, int c, int type, 
	int pref, time_t time, prefix_t *prefix, prefix_t *nexthop, 
	gateway_t *gateway, char *append);
int rib_route_sweep (rib_t * rib, int type);
void delete_all_route_with_type (int type);

void init_rib (trace_t * tr);
int rib_update_route (int cmd, prefix_t *prefix, generic_attr_t *attr, 
		int pref, u_long option);
int add_route_to_rib (int type, prefix_t *prefix, prefix_t *nexthop,
                  interface_t *interface, int pref, u_long opt);
int del_route_from_rib (int type, prefix_t *prefix);
int show_ip_routes (uii_connection_t * uii);
int trace_rib (uii_connection_t * uii, char *s);
int no_trace_rib (uii_connection_t * uii, char *s);

prefix_t *rib_find_best_route (prefix_t *prefix);

extern rib_t *RIB;
#ifdef HAVE_IPV6
extern rib_t *RIBv6;
int show_ipv6_routes (uii_connection_t * uii);
#endif /* HAVE_IPV6 */

#endif /* _RIB_H */
