/*
 * $Id: mrtd.c,v 1.19 1998/06/03 21:40:37 masaki Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/ioctl.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <fcntl.h>
#ifndef SETPGRP_VOID
#include <sys/termios.h>
#endif

#include <mrt.h>
#include <trace.h>
#include <interface.h>
#include <rip.h>
#ifdef HAVE_IPV6
#include <ripng.h>
#endif /* HAVE_IPV6 */
#include <bgp.h>
#include <signal.h>
#include <config_file.h>
#include <select.h>
#include <rib.h>
#include <view.h>

#include "config.h"

rip_t *RIP;
bgp_t *BGP;
#ifdef HAVE_IPV6
ripng_t *RIPNG;
#endif /* HAVE_IPV6 */
int IPV4 = 0;
int IPV6 = 0;


static int
rib_add_rip_route (prefix_t * prefix, rip_attr_t * attr)
{
    return (add_route_to_rib (PROTO_RIP, prefix, attr->nexthop, 0,
			      RIP_PREF, 0));
}


/* add_interfaces
 * add all of the interfaces to the appropriate ribs
 */
static void
add_interfaces_to_rib ()
{
    interface_t *interface;
    ll_addr_t *ll_addr;
    prefix_t *network, *nexthop, *nexthop4;
#ifdef HAVE_IPV6
    prefix_t *nexthop6;

    /* ipv6 dummy nexthop for direct connected route */
    nexthop6 = ascii2prefix (AF_INET6, "::/128");
#endif /* HAVE_IPV6 */
    /* ipv4 dummy nexthop for direct connected route */
    nexthop4 = ascii2prefix (AF_INET, "0.0.0.0/32");

    LL_Iterate (INTERFACE_MASTER->ll_interfaces, interface) {

	if (interface->ll_addr == NULL)
	    continue;
	if (!BIT_TEST (interface->flags, IFF_UP))
	    continue;

	LL_Iterate (interface->ll_addr, ll_addr) {
	    prefix_t *pointopoint = NULL;

	    if (ll_addr->prefix == NULL)
		continue;

	    network = ll_addr->prefix;
	    if (BIT_TEST (interface->flags, IFF_POINTOPOINT))
		pointopoint = ll_addr->broadcast;

	    if (network->family == AF_INET) {
		/* OK */
		nexthop = nexthop4;
	    }
#ifdef HAVE_IPV6
	    else if (network->family == AF_INET6) {

		if (ipv6_compat_addr (prefix_toaddr6 (network))
		    && ll_addr->prefix->bitlen > 0)
		    continue;
		if (ipv6_link_local_addr (prefix_toaddr6 (network)))
		    continue;
		if (ipv6_multicast_addr (prefix_toaddr6 (network)))
		    continue;

		nexthop = nexthop6;
	    }
#endif /* HAVE_IPV6 */
	    else {
		/* skip others */
		continue;
	    }

	    /* create new one for masking */
	    network = New_Prefix (network->family, prefix_tochar (network),
				  network->bitlen);
	    /* masking may not be needed */
	    netmasking (network->family, prefix_tochar (network),
			network->bitlen);

	    trace (TR_INFO, MRT->trace, "Adding an interface route for %s\n",
		   interface->name);
	    add_route_to_rib (PROTO_CONNECTED, network, nexthop, interface,
			      CONNECTED_PREF, 0);
	    Deref_Prefix (network);

	    if (pointopoint) {
		/* create new one for masking */
		pointopoint = New_Prefix (pointopoint->family,
					  prefix_tochar (pointopoint),
					  pointopoint->bitlen);
		/* masking may not be needed */
		netmasking (pointopoint->family, prefix_tochar (pointopoint),
			    pointopoint->bitlen);

		trace (TR_INFO, MRT->trace, 
		       "Adding an interface route for %s\n", interface->name);
		add_route_to_rib (PROTO_CONNECTED, pointopoint, nexthop, 
				  interface, CONNECTED_PREF, 0);
		Deref_Prefix (pointopoint);
	    }
	}
    }
    Deref_Prefix (nexthop4);
#ifdef HAVE_IPV6
    Deref_Prefix (nexthop6);
#endif /* HAVE_IPV6 */
}


/*
 * add all configured static routes to the appropriate rib
 */
static void
add_all_static_to_rib ()
{
    static_route_t *static_route;

    if (CONFIG_MRTD->ll_static_routes == NULL)
	return;

    LL_Iterate (CONFIG_MRTD->ll_static_routes, static_route) {
	add_route_to_rib (PROTO_STATIC, static_route->prefix,
			 static_route->nexthop, static_route->interface, 
			 (static_route->pref >= 0)? static_route->pref: 
						    STATIC_PREF,
			  0);
    }
}


static void
daemonize ()
{
    int pid, t;

    if ((pid = fork ()) == -1) {
	perror ("mrt: fork");
	exit (1);
    }
    else if (pid != 0) {	/* parent */
	exit (0);
    }

#ifdef SETPGRP_VOID
    if ((t = setpgrp ()) < 0) {
	perror ("mrt: setpgrp");
	exit (1);
    }
    signal (SIGHUP, SIG_IGN);

    /* fork again so that not being a process group leader */
    if ((pid = fork ()) == -1) {
	perror ("mrt: fork");
	exit (1);
    }
    else if (pid != 0) {	/* parent */
	exit (0);
    }
#else /* !SETPGRP_VOID */
    if ((t = setpgrp (0, getpid ())) < 0) {
	perror ("mrt: setpgrp");
	exit (1);
    }

    /* Remove our association with a controling tty */
    if ((t = open ("/dev/tty", O_RDWR, 0)) >= 0) {
	if (ioctl (t, TIOCNOTTY, NULL) < 0) {
	    perror ("mrt: TIOCNOTTY");
	    exit (1);
	}
	(void) close (t);
    }
#endif /* SETPGRP_VOID */

#ifdef notdef
    /* Close all open files --- XXX need to check for logfiles */
    for (t = 0; t < 2; t++)
	(void) close (t);
#endif

    umask (022);
}


void
main (int argc, char *argv[])
{
    char c, *p, *name = argv[0];
    extern char *optarg;	/* getopt stuff */
    extern int optind;		/* getopt stuff */
    int errors = 0;
    int daemon = 0;
    char *port = "mrt";
    time_t start = 0, now;

    int kernel_read_flag = 1;
    int kernel_install_flag = 1;
    int rib_install_flag = 1;

    char *usage = "Usage: %s [-f config_file] [-p uii_port ] [-v] [-n]\n";
    char *config_file = NULL;
    trace_t *default_trace;


    if ((p = strrchr (name, '/')) != NULL) {
	name = p + 1;
    }
    if (strcasecmp (name, "mrtd") == 0) {
	config_file = "/etc/mrtd.conf";		/* unix convension */
	daemon = 1;
    }
#ifdef MCHECK
    mcheck (0);
#endif
    default_trace = New_Trace2 ("MRTd");
    set_trace (default_trace, TRACE_PREPEND_STRING, "MRTD", 0);
    set_trace (default_trace, TRACE_MAX_ERRORS, DEFAULT_MAX_ERRORS, 0);

    while ((c = getopt (argc, argv, "rhnkvf:p:")) != -1)
	switch (c) {
	case 'v':		/* verbose */
	    set_trace (default_trace, TRACE_FLAGS, TR_ALL,
		       TRACE_LOGFILE, "stdout",
		       0);
	    daemon = 0;
	    break;
	case 'f':		/* config file */
	    config_file = optarg;
	    break;
	case 'n':		/* no kernel installation */
	    kernel_install_flag = 0;
	    daemon = 0;
	    break;
	case 'k':		/* no kernel read */
	    kernel_read_flag = 0;
	    break;
	case 'r':		/* no rib installation */
	    rib_install_flag = 0;
	    break;
	case 'p':		/* uii port number */
	    port = optarg;
	    break;
	case 'h':
	default:
	    errors++;
	    break;
	}


    if (errors) {
	fprintf (stderr, usage, name);
	printf ("\nMRT %s compiled on %s\n\n",
		MRT_VERSION, __DATE__);
	exit (1);
    }

    if (daemon) {
	if (getuid ()) {
	    fprintf (stderr, "mrt: must be root\n");
	    exit (1);
	}
    }
    /* init_trace (name, daemon); */
    init_trace (name, 1);	/* always syslog */

    init_mrt (default_trace);
    init_mrt_reboot (argc, argv);
    trace (TR_INFO, MRT->trace, "%s compiled on %s started\n",
	   MRT_VERSION, __DATE__);

    /* read information on all interfaces from the kernel */
    init_interfaces (default_trace);
    init_mrtd_config (default_trace);

    init_rip (default_trace);
    /* install the route in the kernel */
    RIP->add_call_fn = rib_add_rip_route;

    init_BGP (default_trace);
    set_BGP (BGP_RECV_UPDATE_FN, process_bgp_update, NULL);
    set_BGP (BGP_STATE_CHANGE_FN, bgp_write_status_change, NULL);
    if (rib_install_flag)
	set_BGP (BGP_RT_UPDATE_FN, rib_update_route, NULL);
    if (INTERFACE_MASTER->default_interface) {
	if (INTERFACE_MASTER->default_interface->primary) {
	    assert (
		INTERFACE_MASTER->default_interface->primary->prefix->family
		       == AF_INET);
	    set_BGP (BGP_MY_ID, prefix_tolong (
		  INTERFACE_MASTER->default_interface->primary->prefix), 0);
	}
    }

    set_uii (UII, UII_PROMPT, 0, "password> ", 0);
    set_uii (UII, UII_PROMPT, 1, "MRT> ", 0);

#ifdef HAVE_IPV6
    ripng_init (default_trace);
    if (rib_install_flag)
	ripng_set (RIPNG_RT_UPDATE_FN, rib_update_route, NULL);
#endif /* HAVE_IPV6 */

    init_rib (default_trace);

    /* default trace flag should be set by here */

    /*
     * read configuration here
     */
    init_config (default_trace);
    /* default_trace may be modified by debug statement in config file */
    if (config_from_file (default_trace, config_file) < 0) {
	UII->initial_state = 1;	/* no password */
	config_create_default ();
    }

    if (daemon) {
	/*
	 * Now going into daemon mode right after reading configuration
	 */
	daemonize ();
    }

    if (BIT_TEST (MRT->protocols, (1 << PROTO_RIP)))
	IPV4++;
    if (BIT_TEST (MRT->protocols, (1 << PROTO_BGP))) {
	if (BGP->num_ipv4_peers)
	    IPV4++;
	if (BGP->num_peers - BGP->num_ipv4_peers > 0)
	    IPV6++ /* BGP4+ */ ;
    }
    if (BIT_TEST (MRT->protocols, (1 << PROTO_RIPNG)))
	IPV6++;

    /* RIP is not supported at the moment -- masaki */
    /* these are inter-proto updates (imported routes) */
    /* At the moment I'm lazy.
       I know Craig uses object oriented aproach, though. */
    if (IPV4) {
        RIB->proto_update_route[PROTO_BGP] = bgp_update_route;
        if (kernel_install_flag)
            RIB->kernel_update_route = kernel_update_route;
    }
#ifdef HAVE_IPV6
    if (IPV6) {
        RIBv6->proto_update_route[PROTO_BGP] = bgp_update_route;
        RIBv6->proto_update_route[PROTO_RIPNG] = ripng_update_route;
        if (kernel_install_flag)
	    RIBv6->kernel_update_route = kernel_update_route;
    }
#endif /* HAVE_IPV6 */

/* CONFIGURATION DEPENDENT INITIALIZATION PART */

    kernel_init ();		/* open routing socket */

    /* add kernel route first, overriden later */
    if (kernel_read_flag) {
	kernel_read_rt_table ();
	time (&start);
    }

    add_interfaces_to_rib ();	/* interface route first */
    add_all_static_to_rib ();	/* static route next */

    if (BIT_TEST (MRT->protocols, (1 << PROTO_RIP)))
	start_rip ();
#ifdef HAVE_IPV6
    if (BIT_TEST (MRT->protocols, (1 << PROTO_RIPNG)))
	ripng_start ();
#endif /* HAVE_IPV6 */
    if (BIT_TEST (MRT->protocols, (1 << PROTO_BGP)))
	start_bgp ();

    listen_uii2 (port);

    running = 1;

    while (1) {
#ifndef HAVE_LIBPTHREAD
	while (process_all_schedules ());	/* process all first */
	mrt_select ();
#else
	sleep (1);
#endif /* THREADS */
	if (start) {
	    time (&now);
#define KERNEL_ROUTE_TIMEOUT 120
	    if (now - start > KERNEL_ROUTE_TIMEOUT) {
		delete_all_route_with_type (PROTO_KERNEL);
		start = 0;
	    }
	}
    }
    exit (0);
}
